
import matplotlib.pyplot as plt
import numpy as np
import scipy.integrate
import scipy.ndimage as ndi

import doas
from . import misc
from . import plume


import scipy.optimize as sopt


def distance(x,p):
    return np.sqrt((x[0] - p[0])**2 + (x[1] - p[1])**2)


def find_min_distance_to_curve(p, c):
    """
    p :: point x0,y0
    c :: polynomial coefficients of 2D curve
    """
    constr = {
        'type': 'eq',
        'fun':  lambda x,c: np.polyval(c,x[0]) - x[1],
        'args': (c,)
    }

    x0 = [p[0], np.polyval(c, p[0])]
    res = sopt.minimize(distance, x0=x0, args=(p,),
            method='SLSQP', constraints=constr,
            options={'maxiter': 100},
    )

    if not res.success:
        print(res)

    return res.x, res.fun



def gauss(p, x,y,u, x0=0.0, y0=0.0):
    """\
        Computes plume in kg/m2.

        x: distance from plume (in km)
        y: distance from center line (in km)

        u: wind speed (in meters)
    """
    # conversion factor (kg/m2 -> ppm)   
    air_column = 1.65e29 # molecules per m2
    m = 44.01e-3         # CO2 molar mass: kg/mol
    NA = 6.02214129e23   # 

    c = 1e6 / m / (air_column / NA)

    # km -> m
    x = x * 1e3 # km -> m
    y = y * 1e3 # km -> m

    # state vector
    Q = p[0] / 0.031536  # emissions in Mt/y (convert to kg/s)
    K = p[1]             # eddy diffusion coefficient (in m^2 / s)
    BG = p[2]            # background field (in ppm)

    # dispersion along the plume
    sigma = np.sqrt(2.0 * K * (x-x0) / u)

    # compute Gaussian plume
    A = Q / (np.sqrt(2.0 * np.pi) * sigma * u)
    B = np.exp(-0.5 * (y-y0)**2 / sigma**2)

    p = A * B
    p[x <= 0.0] = 0.0

    # kg/m2 -> ppm
    p = c*p + BG

    return p



def find_all_roots(hh, dd, d, h_max=5000.0):
    """\
    Find all heights (hh), below hh_max where wind direction (dd) equals
    given wind direction (d).
    """

    founds = []

    for i in range(hh.size-1):
        if dd[i] < dd[i+1]:
            dmin, dmax = dd[i], dd[i+1]
            hmin, hmax = hh[i], hh[i+1]
        else:
            dmin, dmax = dd[i+1], dd[i]
            hmin, hmax = hh[i+1], hh[i]

        if hmin > h_max:
            continue

        if dmin <= d < dmax:
            d0 = np.interp(d, [dmin, dmax], [hmin, hmax])
            founds.append(d0)

    return founds




def arc_length_of_poly(c, a, b):

    x = np.linspace(a,b)
    s = np.sqrt( 1 + np.polyval(np.polyder(c), x)**2 )

    return scipy.integrate.simps(s,x)


def compute_arc_length_and_distance(points, coeff, x_o, x_max):
    """\
    Given n points and polynomial function on [x_o, x_max], compute
    arc length along curve and distance from curve.
    """
    dxs = []
    dys = []

    for px,py in points:
        (qx,qy), dy = find_min_distance_to_curve([px,py], coeff)

        if py > np.polyval(coeff, px):
            dy *= -1.0

        if qx <= x_max:
            dx = arc_length_of_poly(coeff, x_o, qx)
        else:
            dx = np.nan

        dxs.append(dx)
        dys.append(dy)

    return np.array(dxs), np.array(dys)





def flux_inversion(p0, dx, dy, field, wind):

    p0 = np.asarray(p0)

    def f(x,b):
        return gauss(x,*b)

    b = (dx,dy,wind)
    x, res = doas.solver.gauss_newton(p0, field, f, Se=np.nan, b=b)

    return x, res


def setup_flux_inversion(lon0, lat0, field, clouds, is_plume, noise,
        lon_model, lat_model, height_model, uwind, vwind,
        origin=None, lon_o=None, lat_o=None,
        degree=3, epsg=3006,
        **kwargs):
    """
    LEVEL-2 data
    ------------
    lon0, lat0 :: center longitude/latitude of ground pixels
    field      :: satellite observations (e.g. XCO2) - without noise
    clouds     :: cloud fractions at ground pixels
    is_plume   :: boolean array where plume has been detected
    noise      :: noise on field

    MODEL DATA (e.g COSMO)
    ----------------------
    lon_model    :: longitude in model (lon, lat)
    lat_model    :: latitude in model (lon, lat)
    height_model :: height in model (lon, lat, layers+1)
    uwind        :: u-wind in model (lon, lat, layers)
    vwind        :: v-wind in model (lon, lat, layers)

    FLUX INVERSION PARAMETERS
    -------------------------
    origin       :: name of plume (used to get lon_o and lat_o)
    lon_o, lat_o :: longitude/latitude of point source
    degree       :: degree of polynomial fitted on plume

    """
    ##################################
    # DETERMINE CENTER LINE OF PLUME #
    ##################################
    # location of plume
    if origin is not None:
        lon_o, lat_o = plume.get_source_origin(origin)
    i_o, j_o = misc.nearest_index(lon_o, lat_o, lon0, lat0)

    # transform to EPSG system with meters
    lon_o, lat_o = misc.transform_points(lon_o, lat_o, epsg)
    lon0, lat0 = misc.transform_points(lon0, lat0, epsg)
    lon_model, lat_model = misc.transform_points(lon_model, lat_model, epsg)

    # m -> km and origin at point source
    lon0 = (lon0 - lon_o) / 1e3 # km
    lon_model = (lon_model - lon_o) / 1e3 # km
    lon_o = 0.0

    lat0 = (lat0 - lat_o) / 1e3
    lat_model = (lat_model - lat_o) / 1e3
    lat_o = 0.0


    # get location of plume and XCO2 values (plus wind etc.)
    lon_p = lon0[is_plume]
    lat_p= lat0[is_plume]

    # polyfit center of plume (TODO: should work with general 2D curves 
    #(polar coords with source in origin?))
    x = np.linspace(
            min(lon_p.min(), lon_o),
            max(lon_p.max(), lon_o)
    )
    # TODO: weight could/should depend on t-value
    w = np.ones(lon_p.shape)

    # make sure center line goes through point source
    lon_p = np.append(lon_p, lon_o)
    lat_p = np.append(lat_p, lat_o)
    w = np.append(w, 1000.0)

    c = np.polyfit(lon_p,lat_p,degree, w=w)

    # center line is given by x,y
    y = np.polyval(c, x)


    ##################################
    # DISTANCE FROM CENTER LINE      #
    ##################################

    # make plume bigger to include background
    struct = ndi.generate_binary_structure(2, 2)
    full_area = ndi.binary_dilation(is_plume, struct, iterations=3) # TODO iterations?

    # compute for all ground pixels distance to center line
    lon_a = lon0[full_area]
    lat_a = lat0[full_area]
    field_a = field[full_area] + noise[full_area]

    points = np.transpose([lon_a, lat_a])
    dx, dy = compute_arc_length_and_distance(points, c, lon_o, lon_p.max())

    # remove points downstream of lon_p.max()
    lon_a = lon0[full_area][np.isfinite(dx)]
    lat_a = lat0[full_area][np.isfinite(dx)]
    field_a = field[full_area][np.isfinite(dx)] + noise[full_area][np.isfinite(dx)]

    dy = dy[np.isfinite(dx)]
    dx = dx[np.isfinite(dx)]


    ##################################
    # WIND SPEED ALONG CENTER LINE   #
    ##################################

    i_o, j_o = misc.nearest_index(lon_o, lat_o, lon_model, lat_model)

    # estimate cardinal wind direction (and plume height) along center line (TODO)
    uwind = uwind[::-1, i_o, j_o]
    vwind = vwind[::-1, i_o, j_o]
    hh = height_model[::-1, i_o, j_o]
    hh = 0.5 * (hh[1:] + hh[:-1])

    # wind direction profile
    dd = (270.0 - np.rad2deg(np.arctan2(vwind,uwind))) % 360

    # wind direction estimate from center line
    # TODO: slope does only knows westerly winds
    m_o = np.polyval(np.polyder(c), 0.0)
    d_o = (270 - np.rad2deg(np.arctan(m_o))) % 360


    # estimate wind direction (at center line) TODO
    heights = find_all_roots(hh, dd, d_o, h_max=5000.0)
    h_o = heights[0]
    if len(heights) > 1:
        print('Warning: found %d heights with fitting wind direction' % len(heights))

    # compute wind speed at origin (TODO: wind speed along center line)
    u_o = np.interp(h_o, hh, uwind)
    v_o = np.interp(h_o, hh, vwind)
    wind = np.sqrt(u_o**2 + v_o**2)


    # TODO: mask cloudy pixels

    return x,y, lon_a, lat_a, field_a, c, dx, dy, wind


    # TODO: distance and wind speed can be used to determine age of plume













